"""
Ten skrypt analizuje dane dotyczące artykułów naukowych z bazy danych arXiv, a następnie generuje nowe tytuły artykułów dla autorów, korzystając z usługi OpenAI.
Skrypt składa się z wielu funkcji, które odpowiednio przetwarzają dane wejściowe, generują tytuły i zapisują wyniki.
"""

import itertools
import json
import pickle as pkl
import re
import string
from collections import defaultdict
from typing import List
import matplotlib.pyplot as plt
import numpy as np
import openai
import pandas as pd
from adjustText import adjust_text
from pandas import DataFrame
from tqdm import tqdm

ARXIV_METADATA_FILE_PATH = './data/arxiv_metadata_2022.pkl'

query_settings = {
    "papers_in_context": 5,
    "minimum_author_papers": 2,
    "prompt_start": 'Here is a list of related machine-learning papers:\n\n> '
}

openai_completion_settings = {
    "completions_per_prompt": 5,
    "engine": "text-davinci-002",
    "stop": ">"
}

openai_api_settings = {
    "api_key": "insert - your- api - key ",
}


# Utility
def open_pickled_file(path: string, open_with_pandas=False) -> "DataFrame":
    file = open(path, 'rb')
    return pd.read_pickle(file) if open_with_pandas else pkl.load(file)


def save_date_to_pickle(data, path: string):
    pkl.dump(data, open(path, 'wb'))


def contains_num(x):
    if True in [char.isdigit() for char in x]:
        return True
    return False
# End Utility


def compute_author_stats(dataframe_from_arxiv_metadata: DataFrame) -> tuple[dict, list]:
    """
    Oblicza statystyki dotyczące autorów na podstawie podanych metadanych z arXiv.
    :param dataframe_from_arxiv_metadata: DataFrame zawierający metadane artykułów z arXiv.
    :return: Krotka zawierająca słownik z informacją o autorach i indeksach ich artykułów oraz posortowaną listę autorów.
    """
    def get_authors_list(s: str) -> List:
        return [
            a.lower().strip()
            for a in re.split('and |, |\n', s)
        ]

    authors_list_arr = dataframe_from_arxiv_metadata.authors.apply(get_authors_list)
    authors_dict = defaultdict(list)
    for i, l in enumerate(authors_list_arr):
        for author in l:
            authors_dict[author].append(i)

    pruned_authors_list = sorted(set(list(itertools.chain.from_iterable(authors_list_arr))))

    return authors_dict, pruned_authors_list


def sanitize_author_collections(authors_dict: dict, authors_list: list):
    """
    Usuwa niepoprawne dane (dane, które nie są autorami) z kolekcji autorów.
    :param authors_dict: Słownik autorów.
    :param authors_list: Lista autorów.
    :return: Krotka zawierająca oczyszczone dane słownika i listy autorów.
    """

    keywords = re.compile(r'universi|departm|faculty|institute|global|center|laboratory')

    def author_has_more_articles_than(author, number_of_articles):
        return len(author) > number_of_articles

    def does_not_contain_punctuaction(author):
        return not any(punctuaction in author for punctuaction in string.punctuation)

    def author_has_name_and_surname(author):
        return len(author.split()) > 1 and len(author.split()[0]) > 1

    def author_does_not_contain_keywords(author, keywords):
        return keywords.search(author) is None

    authors_list_clean = [
        author for author in authors_list
        if author_does_not_contain_keywords(author, keywords)
           and does_not_contain_punctuaction(author)
           and not contains_num(author)
           and author_has_more_articles_than(author, 3)
           and author_has_name_and_surname(author)
    ]
    authors_dict_clean = {
        author: authors_dict[author]
        for author in authors_list_clean
    }
    assert len(authors_list_clean) == len(authors_dict_clean)

    return authors_dict_clean, authors_list_clean


def get_author_paper_counts(authors_dict: dict):
    paper_idxs = authors_dict.values()
    return np.array([len(x) for x in paper_idxs])


def create_authors_plot(authors_paper_count, authors_list: list):
    """
    Tworzy wykres ilustrujący liczbę autorów w zależności od liczby opublikowanych artykułów.
    :param authors_paper_count: Tablica z liczbą artykułów dla każdego autora.
    :param authors_list: Lista autorów.
    """
    plt.figure(figsize=(5, 8))
    plt.xscale('log')
    count, count_freq = np.unique(authors_paper_count, return_counts=True)
    # plt.bar(x=count, height=count_freq)
    plt.barh(y=count, width=count_freq)
    plt.xlabel('Number of authors')
    plt.ylabel('Number of arXiv papers')

    # get top20 authors
    ind = np.argpartition(authors_paper_count, -20)[-20:]
    ind = ind[np.argsort(authors_paper_count[ind])][::-1]
    texts = []
    for offset, i in enumerate(ind):
        texts.append(plt.text(y=authors_paper_count[i], x=10, s=authors_list[i] + f' ({authors_paper_count[i]})'))
    adjust_text(texts, autoalign='y')
    plt.show()


def get_authors_titles_dict(authors_paper_count, authors_dict, dataframe_from_arxiv_metadata):
    """
    Tworzy słownik, w którym kluczem jest autor, a wartością lista tytułów artykułów.
    W pierwszym kroku pobiera tytuły artykułów z DataFrame dataframe_from_arxiv_metadata.
    Następnie tworzy słownik authors_dict_titles, przypisując każdemu autorowi odpowiednie tytuły artykułów.
    Na końcu sortuje autorów według liczby artykułów i zwraca słownik z posortowanymi danymi.
    :param authors_paper_count: Tablica z liczbą artykułów dla każdego autora.
    :param authors_dict: Słownik autorów.
    :param dataframe_from_arxiv_metadata: DataFrame z metadanymi z arXiv.
    :return: Słownik, w którym kluczem jest autor, a wartością lista tytułów artykułów.
    """

    titles = dataframe_from_arxiv_metadata['title'].values
    authors_dict_titles = {
        author: titles[authors_dict[author]]
        for author in tqdm(authors_dict.keys())
    }

    # sort by num papers and save
    authors = np.array(list(authors_dict_titles.keys()))
    args = np.argsort(authors_paper_count)[::-1]
    return {
        a: authors_dict_titles[a]
        for a in authors[args]
    }


def prepare_authors_titles_dict(data_path):
    df = open_pickled_file(data_path, open_with_pandas=True)
    (authors_dict, authors_list) = compute_author_stats(df)
    (authors_dict_clean, authors_list_clean) = sanitize_author_collections(authors_dict, authors_list)
    authors_paper_count = get_author_paper_counts(authors_dict_clean)
    #create_authors_plot(authors_paper_count, authors_list_clean)
    return get_authors_titles_dict(authors_paper_count, authors_dict_clean, df)


def configure_openai_api(settings: dict):
    print("\nConfiguring OpenAI API...")
    assert "api_key" in settings
    openai.api_key = settings["api_key"]
    print("\nConfigured OpenAI API!")


def create_openai_completion(prompt, settings):
    print(f"\nCreating OpenAI Completion with prompt... [prompt={prompt}]")
    completion = openai.Completion.create(
        engine=settings["engine"], prompt=prompt,
        n=settings["completions_per_prompt"], stop=settings["stop"]
    )
    print(
        f"\nSuccessfully completed OpenAI Completion! [completions={[f'{i + 1}: {completion.choices[i].text},' for i in range(len(completion.choices))]}")
    return completion


def create_prompt(author, authors_titles_dict, settings, stop):
    return settings["prompt_start"] + f'\n{stop} '.join(
        authors_titles_dict[author][-settings["papers_in_context"]:]) + f'\n{stop}'


def filter_authors(authors, predicate):
    return [author for author in authors if predicate(author)]


def save_results(data, path, pickled=False):
    pkl.dump(data, open(path, 'wb')) if pickled else json.dump(data, open(path, 'w'), indent=2)


def generate_authors_titles(authors_titles_dict, author_filter, iteration_limit=-1, progress_bar=False, path=None):
    assert author_filter
    authors_save = {}
    authors = list(authors_titles_dict.keys())
    filtered_authors = filter_authors(authors, author_filter)
    authors_for_iteration = enumerate(tqdm(filtered_authors)) if progress_bar else enumerate(filtered_authors)
    configure_openai_api(openai_api_settings)
    for i, author in authors_for_iteration:
        if iteration_limit != -1 and i + 1 > iteration_limit:
            break
        if author not in authors_save:
            prompt = create_prompt(author, authors_titles_dict, query_settings, openai_completion_settings["stop"])
            completion = create_openai_completion(prompt, openai_completion_settings)
            authors_save[author] = [completion.choices[i].text for i in range(len(completion.choices))]
    if path is not None:
        save_results(authors_save, path)
    return authors_save


def print_authors_titles_given_authors(authors, data=None, path=None, read_pickled=False):
    if data is None and path is None:
        print("no data or path provided")
        return

    def _print(authors, authors_save_processed):
        for author in authors:
            print(author)
            titles = authors_save_processed[author]

            for title in titles:
                print('\t', title)

    if path is not None:
        _print(authors, pkl.load(open(path, 'rb'))) if read_pickled else _print(authors, pd.read_json(path))
        return

    _print(authors, data)


def print_authors_titles_head(n, data=None, path=None, read_pickled=False):
    if data is None and path is None:
        print("no data or path provided")
        return

    if path is not None:
        generated_data = pkl.load(open(path, 'rb')) if read_pickled else pd.read_json(path)
        print_authors_titles_given_authors(generated_data.keys()[:n], generated_data)
        return

    if data is not None:
        authors = data.keys()[:n]
        print_authors_titles_given_authors(authors, data)


if __name__ == '__main__':
    authors_dict = prepare_authors_titles_dict(ARXIV_METADATA_FILE_PATH)
    generated_data_path = 'authors_save_full.json'
    generated_data = generate_authors_titles(authors_dict,
                                             lambda author: len(authors_dict[author]) > 2,
                                             iteration_limit=10,
                                             path=generated_data_path,
                                             progress_bar=True)
    print_authors_titles_head(5, path=generated_data_path)